#!/usr/bin/python
# -*- coding: UTF-8 -*-
#    02.12.2013

# Hier das Grundgerüst einer Anwendung dargestellt:
# 1.) Es wird eine Instanz von Xbee() angelegt.
# 2.) Eine Endlosschleife beginnt und startet mit einem sleep.
# 3.) Nach Ablauf des Sleep
# 4.) - werden Unterprogramme ausgeführt, die zwischenzeitlich in eine ToDo-Liste
#       eingefügt wurden.
# 5.) - werden andere Unterprogramme in regelmäßigen Zeitabschnitten ausgeführt.
#
# 6.) hier wird am lokalen oder an einem remote-Device der Pin P0 (RSSI)
#     alle 1 Sekunden umgeschaltet.
#     Wenn das lokale Device der Coordinator ist,
#     dann kann der per RemoteAtCmd mit dev = 0 adressiert werden
#
#     Zuerst wird mit ATPO4 der Pin aus digitaler Ausgang geschaltet
#     Dann wird jede Sekunde ein ATP0 gesendet.
#     Die Antwort lautet dann ATP04 oder ATP05 und wird von togglePin()
#     ausgewertet.
#     togglePin() ruft dann self.addToDo() auf und übergibt ein Tuple mit
#     einer codierten Anweisung, die in die ToDo-Liste eingetragen wird.
#     Die ToDo-Liste kurzen Abständen auf Einträge überprüft,
#     wenn einer vorhanden ist, dann wird ausgewertet und ein Aktion ausgelöst
#     - in diesem Falle wird Frame abgesandt, das den Pin umschaltet.

#     Das Konzept mit der todolist[] ist zwischenzeitlich entfallen.
#     Anstatt die Jobs erst in eine Liste einzutragen und dann ausführen zu
#     lassen, werden sie direkt verschickt.

import atexit
import time
import xbee
import SerPort
import xbee_help

class App(object):
    def __init__(self):
        self.ser = SerPort.open()
        self.lock = False
        self.todolist = []
        self.start()


    def addToDo(self, data):
        """ Fügt ein Tuple in die ToDo-Liste ein """
        while self.lock:
            time.sleep(.01)
        self.lock = True
        self.todolist.append(data)
        self.lock = False


    def receive(self, station, frame):
        """ Wird als Callback-Funktion an das Objekt Xbee() übergeben und
        liefert folgende Daten zurück:
        - station = Index des Absenders in der Senderliste,
        - frame = komplettes Frame als Bytefolge (ohne abschließende Prüfsumme).
        Die Frames 0x88, 0x8B und 0x97 werden im Normalfall vom rxHandler
        interpretiert und kommen hier nicht an !
        """
        # die Daten zur Interpretation an xbee_help weiterreichen
        xbee_help.x_print(frame)


    def togglePin(self, station, frame):
        """ Schaltet einen IO-Pin eines XBee um,
            Es wird geprüft, ob das Lokale Modem oder ein remote Modem gemeint ist.
            Das aufrufende Programm legt mit der Abfrage eines Pins fest,
            welcher Pin getoggled werden soll.
            Der muss allerdings als Ausgang geschaltet sein:
            wenn 4 gelesen wird, wird 5 geschrieben
            und umgekehrt.
        """

        if frame[3] == 0x97 and len(frame) == 19:    # frame0x97 mit 1 Datenbyte
            out = frame[18]
            dev = self.X.getStation(adr16 = frame[13:15])[0]
            if out == 4:
                out = (5,)
            elif out == 5:
                out = (4,)
            else:
                out = (out,)
            # Format: Action, Frame, station, atcmd, data
            self.addToDo( (1, 0x17, dev, frame[15:17].decode(), out) )

        elif frame[3] == 0x88 and len(frame) == 9:  #frame 0x88 + 1 Datenbyte
            out = frame[8]
            dev = self.X.getStation(adr16 = frame[13:15])[0]
            if out == 4:
                out = (5,)
            elif out == 5:
                out = (4,)
            else:
                out = (out,)
            # Format: Action, Frame, station, atcmd, data
            self.addToDo( (1, 0x08, None, frame[5:7].decode(), out) )

        else:
            print("togglePin() -> Fehler :", frame)


    def start(self):
        """ Die Hauptschleife der Applikation """
        self.X = xbee.Xbee(ser = self.ser, callback = self.receive, readCTS = True)
        atexit.register(self.X.exit)

        # Hier wird P0 am Coordinator als Ausgang geschaltet:
        self.X.sendRemoteAtCmd(dev = 0, data = (4,), atcmd = "P0")
        #self.X.sendLocalAtCmd(atcmd = "P0", data = (4,))

        lasttime = int(time.time())
        try:
            while True:
                time.sleep(.1)

                # wenn Jobs in der todo-Liste eingetragen sind
                if len(self.todolist):
                    self.todo()

                # in regelmäßigen Abständen (1 Sekunde) Unterprogramme ausführen
                if int(time.time()) > lasttime:
                    lasttime = int(time.time())
                    self.cronjob(lasttime)  # die aktuelle Zeit wird übergeben

        except KeyboardInterrupt:
            self.X.exit()
            exit()

        self.X.exit()
        exit()


    def cronjob(self, t):
        """ führt Jobs in immer gleichen Zeitintervallen aus.
            Wird jede Sekunde 1-mal aufgerufen und prüft,
            ob es Zeit ist, einen Job zur Ausführung zu bringen.
        """
        # alle 5 Sekunden gemäß interner Uhr
        if t % 5 == 0:
             self.X.sendRemoteAtCmd(dev = 0, atcmd = "P0", rxHandler = self.togglePin)
            #self.X.sendLocalAtCmd(atcmd = "P0", rxHandler = self.togglePin)


        # eine volle Minute (+ 10 Sekunden) gemäß interner Uhr
        # die Jobliste ausmisten, die älter als 180 Sekunden sind ...
##        if t % 60 == 10:
##            self.X.cleanJobs()

        # eine volle Minute (+ 15 Sekunden) gemäß interner Uhr
##        if t % 60 == 15:
##            pass

        # alle 3 Minuten + 17 Sekunden gemäß interner Uhr
##        if t % 180 == 17:


        # volle Stunde (+3 Sekunden) gemäß interner Uhr
##        if t % 3600 == 3:
##            pass


    def todo(self):
        # Alle Jobs, die das Hauptprogramm ausführen soll,
        # werden in einer ToDoListe abgelegt.
        # Hier wird immer nur der erste Datensatz ausgeführt und dann gelöscht.
        # Damit rücken eventuelle nachfolgende Jobs an den Anfang der Liste
        # und werden nach dem nächsten Sleep ausgeführt.

        # überholt, dieser Ansatz wurde nicht weiterverfolgt, aber er funktioniert
        # Action == 1:
        # - Damit können AtCmd's ausgeführt oder Datenframes verschickt werden
        # - Format: Action, Frame, Zieldevice, AtCmd, Datenliste

        #print(self.todolist[0])

        if self.todolist[0][0] == 1:
            if self.todolist[0][1] == 0x08:     # Local AT Command
                self.X.sendLocalAtCmd(atcmd = self.todolist[0][3],
                                        data = self.todolist[0][4])

            elif self.todolist[0][1] == 0x10:   # DataFrame
                self.X.sendDataFrame(dev = self.todolist[0][2],
                                        data = self.todolist[0][4] )

            elif self.todolist[0][1] == 0x17:   # Remote AT Command
                self.X.sendRemoteAtCmd(dev = self.todolist[0][2],
                                        atcmd = self.todolist[0][3],
                                        data = self.todolist[0][4])
            else:
                print ("todo() -> ", self.todolist[0], len(self.todolist))

            while self.lock == True:
                time.sleep(.01)
            self.lock = True
            del self.todolist[0]
            self.lock = False

if __name__ == "__main__":
    App()

# --- [ eof ] ------------------------------------------------------------------