# -*- coding: UTF-8 -*-
#       24.11.2013
#
# --- [ xbee.py ] ----------------------------------------------------------
#
# Ist zuständig für das Versenden von Frames,
# dazu gibt es die drei wesentlichen Methoden:
# -> sendLocalAtCmd(atcmd, data [,frameId, rxHandler])
# -> sendRemoteAtCmd(device, atcmd, data [,frameId, rxHandler, RemoteCmdOptions])
# -> sendDataFrame(device, data [,frameId, rxHandler, BroadcastRadius, txRequestOptions])
# alle Übertragungen können mit einem Zeiteintrag in eine Logdatei eingefügt werden.
#
# Die empfangenen Daten werden per Callback an main() zurückgeliefert und enthalten:
# - Index des Senders in der Senderliste
# - Daten als Bytefolge (ohne Prüfsumme)

import time
import xbee_adr
import xbee_rx
import xbee_jobs

class Xbee(object):
    def __init__(self,
                ser,
                callback  = None,
                readRSSI  = False,
                logToTTY  = False,
                logToFile = False,
                filename  = None,
                readCTS   = False
                ):
        self.ser       = ser
        self.callback  = callback
        self.logToFile = False
        self.logToTTY  = logToTTY
        self.logToFile = logToFile
        self.readRSSI  = readRSSI
        self.readCTS   = readCTS
        self.RSSIstate = True
        self.tm_dez    = True       # Zeit mit 1/10-Sekunden
        self.filehandle= None
        self.cnt_tx    = 0          # zählt die gesendeten remoteAT und DataFrames
        self.cnt_rx    = 0          # zählt die Antworten dazu
        self.cnt_err   = 0          # zählt die Übertragungsfehler dabei
        self.cnt_data  = 0          # zählt die empfangenen DataFrames

        self.adr64bit = bytes(8)    # eigene 64BitSeriennummer
        self.name  = "?"            # eigener Name
        self.id    = None

        # Für jeden Frametyp wird eine eigene ID geführt.
        # Die Werte werden zwischen 1 und 255 incrementiert, '0' entfällt !
        # Eine ID von 0 bedeutet, dass der Empfänger der Nachricht keine Antwort
        # als Bestätigung senden wird.
        # Der Wert wird vor jeder Sendung incrementiert.
        self.frameId08 = 0
        self.frameId10 = 0
        self.frameId17 = 0
        self.frameIdDB = 0

        if type(filename) == str:
            # wird ein String übergeben, dann wird der als Dateiname gewählt,
            # sofern er einen Inhalt hat.
            # Bei einem leeren String wird ein Dateiname generiert.
            # Die Vorgabe ist None - es wird keine LogFile angelegt
            self.filehandle = self.__openLogFile(filename)

        # Einen Thread starten, der für den Empfang zuständig ist,
        # der muss natürlich auf die Serielle Schnittstelle zugreifen
        # und die empfangenen Daten zurückgeben können.
        self.X_rx = xbee_rx.Xbee_rx(ser = self.ser, callback = self.__getReceivedData)
        self.X_rx.start()

        # Ein Objekt zur Verwaltung der StationsAdressen anlegen
        # Dort wird die Liste der Stationen aus einer Datei eingelesen
        self.X_adr = xbee_adr.Xbee_adr(filehandle = self.filehandle,
                    logToFile = self.logToFile, logToTTY = self.logToTTY)

        # Klasse zur Verwaltung von Nachrichten
        self.X_jobs = xbee_jobs.Xbee_jobs()





    def exit(self):
        """ Beendet den Hintergrundthread
            und schließt ggf. noch offene Dateien
        """
        #self.X_adr.getStationList()
        if self.filehandle is not None:
            self.filehandle.write("close(): " +
                                    time.strftime("%d.%m.%Y %H:%M:%S") + "\n")
            self.filehandle.close()
            print("closing file")
            self.filehandle = None
        self.X_rx.exit()                  # den Thread rx stoppen
        print("close Thread RX ... ", end = "")


    def __doNix(self, station, frame):
        """ Hier laufen alle eingehenden Frames auf, für die kein
            spezieller rxHandler definiert wurde.
            Wenn frame[0] = 0x7E, dann muss der cmdstatus geprüft werden,
            ob ein Übertragungsfehler vorliegt.
            wenn frame[0] = 0xA5, dann wurde der Eintrag wegen Zeitüberschreitung
            aus dem Array jobs[] gelöscht
            wenn frame[0 = 0x5A, dann gab es mehrere identische Einträge
            im Array jobs[], ein Überlauf der FrameId muss sich ereignet haben,
            dabei müssen auch mehrere Fehlermeldung ausgegeben werden !
        """
        #return

        msg = "__doNix(): "
        status = 0

        if frame[0] == 0x7E:
            if frame[3] == 0x88:
                status = frame[7]
            elif frame[3] == 0x8B:
                status = frame[8]
            elif frame[3] == 0x97:
                status = frame[17]

            if status:
                msg += "FrameErr, Status: {:02X}, Frame:".format(status)
            else:
                msg += "Frame OK:"

            for i in frame:
                msg += " {:02X}".format(i)

        elif frame[0] == 0xA5:              # 0xA5, FrameId, Frametyp, Station
            msg += "Zeitüberschreitung für FrameId {}, FrameTyp {}, Station: {}".\
                            format(frame[1], frame[2], station)

        elif frame[0] == 0x5A:              # 0x5A, FrameId, Frametyp, Station'
            msg += "mehrfacher Eintrag für FrameId {}, FrameTyp {}, Station: {}".\
                                        format(frame[1], frame[2], station)
        else:
            msg += "error"

        if status > 0:
            print(msg)


    def __doGarNix(self, station, frame):
        """ hier unterbleibt jegliche Ausgabe """
        pass


    def __getReceivedData(self, databytes, timestring):
        """ Holt als Callback aus dem EmpfangsThread die Daten ab:
            databytes  = empfangene Daten als Bytestring ohne Prüfsumme am Ende
            timestring = Textstring mit dem Zeitpunkt des Empfangs
        """
        idx = -1                    # Absender unbekannt oder lokales AT-Frame
        frametyp = databytes[3]     # hier steht immer der Frametyp

        # Die Antwort-Frames auf Local AT, Remote AT und DataFrames liefern einen
        # CommandStatus zurück, der auf Übertragungsfehler hinweist,
        # wenn der Wert > 0 ist
        # Weiterhin führen diese Frames die FrameId der initialisierenden Sendung
        # mit - damit können "Frage" und "Antwort" einander zugeordnet werden.
        cmdstatus = 0
        frameId   = -1              # für alle Frames, die keine FrameId liefern

        if frametyp == 0x88:
            cmdstatus = databytes[7]
            frameId   = databytes[4]

        elif frametyp == 0x8B:
            self.cnt_rx += 1
            cmdstatus = databytes[8]
            frameId   = databytes[4]

        elif frametyp == 0x97:
            self.cnt_rx += 1
            cmdstatus = databytes[17]
            frameId   = databytes[4]

        # --- Bestimmung der Empfangsstärke ------------------------------------

        # Wenn ein Frame einer anderen Station empfangen worden ist
        # und readRSSI == True,
        # dann wird versucht, die Empfangsstärke des Signals zu ermitteln.
        # Dazu wird das lokale AT-Kommando "DB" gesendet.
        # Die Antwort geht als Frame0x88 ein.

        # Wenn allerdings Frames in schneller Folge empfangen werden,
        # dann wird möglicherweise die Empfangsstärke des nächsten Frame geliefert.
        # Dann würde die Signalstärke u.U. einer falschen Station zugeordnet.
        # Das Problem kann nicht grundsätzlich vermieden werden,
        # aber mit Hilfe von RSSIstate wird es versucht:
        #
        # RSSIstate = True  - kein AT-DB in Arbeit
        # RSSIstate = Tuple - (FrameId, SenderId) ein AT-DB ist abgeschickt worden
        # RSSIstate = False - ein zusätzliches Frame ist während eines laufenden
        #                     AT-DB eingetroffen, um eine falsche Zuordnung zu
        #                     vermeiden ist das laufende AT-DB zu canceln

        # ein Sonderfall zur Ermittlung der letzten Empfangsstärke via "AT DB"
        if  frametyp == 0x88 \
                and self.readRSSI \
                and databytes[5:7] == bytes("DB", "ascii")\
                and cmdstatus == 0:
            # Das Frame 0x88 liefert zwar keine Adressdaten, aber es liefert
            # nach einem ATDB die Empfangsstärke der letzten Übertragung !
            # Aber nur dann, wenn readRSSI == True und kein Übertragungsfehler
            # RSSIstate ist:
            # False - wenn ein ein AT DB gecancelt werden muss
            # Tuple - (FrameId und Senderkennung)
            # True  - wenn ein "normales" AT DB versandt wurde
            if self.RSSIstate == False: # abbrechen und beenden
##                print("__getReceivedData: RSSIstate = False")
                self.RSSIstate = True
                return

            elif type(self.RSSIstate) == tuple:
                # wenn die FrameIds von Sendung und Empfang übereinstimmen.
##                print("__getReceivedData: RSSIstate:", self.RSSIstate)
                # noch eine zweite Sicherung:
                # es darf zwischenzeitlich nur 1 einziges Frame empfangen worden
                # sein - nämlich diese Antwort auf das lokale AT DB,
                # sind witere Frames eingegangen, dann wird möglicherweise
                # Empfangstärke einer anderen Station gemessen.
                if self.RSSIstate[0] == databytes[4]:
                    # den DB-Wert in die Stationsliste eintragen
                    if ((self.RSSIstate[1] + 1) & 0xFF) == self.X_rx.framecnt:
##                        print(self.RSSIstate[2], self.X_rx.framecnt)
                        self.X_adr.setATDB(self.RSSIstate[1], ord(databytes[8:9]))
                # RSSIstate immer zurücksetzen, auch keine Übereinstimmung vorhanden
                self.RSSIstate = True
                return
            # bei RSSIstate == True gehts weiter zur Auswertung...

        elif frametyp < 0x8B:
            # die Frames 0x88 und 0x8A enthalten keine
            # für die Adressauflösung / Empfangsstärke auswertbaren Informationen
            pass

        elif frametyp == 0x8B:
            # die Antwort auf ein Frame0x10 liefert nur eine 16Bit Adresse, keine
            # 64-Bit Adresse, die direkt zur Bestimmung der Empfangsstärke genutzt
            # werden kann.
            # Es wird versucht, die 16-Bit Adresse in der Stationsliste
            # zu finden (sofern bereits dort eingetragen ist).
            # Wird die 16-Bit Adresse nicht in der Stationsliste gefunden,
            # dann wird -1 zurückgegeben, sonst der Index.
            idx = self.X_adr.getStation(adr16 = databytes[5:7])[0]

            if cmdstatus == 0:      # keine Übertragungsfehler
            # die Einträge mit Index 0 und 1 werden nicht durchsucht.
                if idx > 1:
                    if self.readRSSI:
                        if self.RSSIstate is True:
                            self.RSSIstate = (self.__sendATDB(), idx)
                        else:
                            self.RSSIstate = False
            else:                   # ein Übertragungsfehler wurde erkannt
                # den Zähler für Fehler incrementieren
                self.X_adr.incrErrCnt(idx)

        # --- StationsId bestimmen ---------------------------------------------

        # Alle anderen Frames enthalten die 64-Bit und 16-Bit Adresse des Absenders.
        # Die 16-Bit Adresse wird vom Coordinator vergeben und ist
        # beim Start des Netzwerkes als Vorgabe 'FFFE'.
        #elif frametyp in (0x90, 0x91, 0x92, 0x94, 0x95, 0x97, 0xA0, 0xA3):
        else:
            if frametyp == 0x97:
                # hier verschiebt sich die Adresse um 1 Byte wegen der FrameId
                adr64 = databytes[5:13]
                adr16 = databytes[13:15]

            else:
                self.cnt_data += 1
                adr64 = databytes[4:12]
                adr16 = databytes[12:14]
            # hier wird die 64-Bit Adresse gesucht und anschließend die
            # gespeicherte 16-Bit Adresse mit der empfangenen verglichen.
            # idx liefert den Index des Senders innerhalb der Senderliste
            idx = self.X_adr.checkAdr64(adr64, adr16)
            # die Anzahl der empfangenen Sendungen incrementieren
            if cmdstatus == 0:
                self.X_adr.incrRxCnt(idx)
            # die Anzahl der Übertragungsfehler mitzählen
            elif cmdstatus > 0:
                self.cnt_err += 1           # Fehler allgemein zählen
                self.X_adr.incrErrCnt(idx)  # Fehler stationsabhängig eintragen

            # wenn readRSSI == True die Empfangsstärke des letzten Signal auswerten
            if self.readRSSI and cmdstatus == 0:
                # sendATDB liefert -> frameID
                # Ein Tuple anlegen mit FrameId und Senderkennung
                # Wenn bereits ein Eintrag besteht, dann ist die letzte Abfrage
                # per Frame0x88 noch nicht abgearbeitet.
                # Um eine fehlerhafte Zuordnung zu vermeiden,
                # wird RSSIstate auf False gesetzt, die Auswertung wird verworfen.
                # wenn RSSIstate = True, dann ist keine Auswertung unterwegs
                # und eine neue Auswertung kann gestartet werden.
                if self.RSSIstate is True:
                    self.RSSIstate = (self.__sendATDB(), idx, self.X_rx.framecnt)

                else:
                    self.RSSIstate = False

        # --- Auswertung der Daten ---------------------------------------------

        # Bei diesen Frames prüfen, ob eine Funktion hinterlegt ist,
        # die beim Empfang automatisch ausgeführt werden soll.
        if frametyp in (0x88, 0x8B, 0x97):
            # Übergabe: frameId, frameTyp
            i = self.X_jobs.checkJobs(frameId, frametyp)
            if i >= 0:
                # Übergabe: Index in der Jobliste, frame
                self.X_jobs.doJob(i, databytes)

            else:
                # wenn keine Funktion hinterlegt war
                # dann zur Bearbeitung nach main() zurückreichen
                if self.callback is not None:
                    self.callback(idx, databytes)
        # alle anderen Frames immer nach main() zurückgereicht
        else:
            if self.callback is not None:
                self.callback(idx, databytes)

        # --- es folgt ggf. noch Ausgabe einzeiliger Logdaten ------------------

        if (self.logToTTY + self.logToFile):
            string = "-> {} ".format(timestring)
            for i in databytes:
                string += "{:>02X} ".format(i)

            if self.logToTTY:
                print(string)

            if self.logToFile and self.filehandle is not None:
               self.filehandle.write(string + "\n")


# --- FRAME0x08 --- Local AT ---------------------------------------------------
    def sendLocalAtCmd(self,
                    atcmd,                 # erforderlich
                    data       = [],       # optionaler Parameter
                    rxHandler  = None,
                    setFrameId = True      # wenn False wird frameId = 0
                    ):
        # Rückgabe: keine
        if rxHandler is None:
            rxHandler = self.__doNix

        cmd = bytearray([0x7E, 0, 0, 0x08]) # Start Delimiter, Länge, Frame-Typ

        if setFrameId:                  # Frame-ID fortlaufend
            self.frameId08 += 1
            if self.frameId08 > 255:
                self.frameId08 = 1
            cmd.append(self.frameId08)

        else:
            cmd.append(0)

        atcmd = atcmd.upper()           # AT-Command in Großschreibung
        cmd.append(ord(atcmd[0]))       # ATCMD
        cmd.append(ord(atcmd[1]))       # ATCMD

        for i in data:
            cmd.append(i)

        l, c = self.__check_sum(cmd)    # Checksum und Länge bestimmen und anhängen
        cmd[2] = l                      # Länge im dritten Byte einfügen
        cmd.append(c)                   # Checksumme anhängen

        if not self.ser is None:
            if self.readCTS and not self.ser.getCTS():
                time.sleep(0.01)
            self.ser.write(cmd)         # senden
#           self.cnt_tx += 1

        # ggf noch einzeilige Druckausgabe
        if (self.logToFile + self.logToTTY):
            t = "<- {} ".format(self.__format_time())
            for i in cmd:
                t+= "{:>02X} ".format(i)

            if self.logToTTY:
                print(t)

            if self.logToFile and self.filehandle is not None:
                self.filehandle.write(t + "\n")

        # einen Eintrag in der Job-Liste vornehmen
        if setFrameId:
            self.X_jobs.addJob(self.frameId08, 0x88, -1, rxHandler)


# --- FRAME0x17 --- Remote AT --------------------------------------------------
    def sendRemoteAtCmd(self,
                dev ,
                atcmd,                  # erforderlich
                data   = [],            # optionale Parameter
                rxHandler = None,       # Function zur Ausführung beim Empfang
                setFrameId = True,      # wenn False, dann frameId = 0
                RemoteCmdOptions = 2    # Options : 2 = Apply Changes
                ):
        # Rückgabe: keine
        if rxHandler is None:
            rxHandler = self.__doNix

        # Start Delimter, Platzhalter für die Länge, Frame-Typ
        cmd = bytearray([0x7E, 0, 0, 0x17])

        if setFrameId:
            self.frameId17 += 1
            if self.frameId17 > 255:
                self.frameId17 = 1
            cmd.append(self.frameId17)

        else:
            cmd.append(0)

        if type(dev) is int:
            adr = self.X_adr.getStation(idx = dev)[1:3]
        # liefert auch den Index zurück
        elif type(dev) is str:
            tmp = self.X_adr.getStation(name = dev)
            dev = tmp[0]
            adr = tmp[1:3]
        else:
            print("Error typ dev")

        # incrementiert den Zähler für gesendete Frames
        self.X_adr.incrTxCnt(dev)

        adr_64bit = ""
        for i in range(8):              # die 64-bit Destination Adresse
            adr_64bit += "{:>02X}".format(adr[0][i])
            cmd.append(adr[0][i])

        adr_16bit = ""
        for i in range(2):              # die 16-bit Destination Network Adresse
            adr_16bit += "{:>02X}".format(adr[1][i])
            cmd.append(adr[1][i])

        atcmd = atcmd.upper()
        cmd.append(RemoteCmdOptions)    # Options : 2 = Apply Changes
        cmd.append(ord(atcmd[0]))       # ATCMD
        cmd.append(ord(atcmd[1]))       # ATCMD

        datastr = ""                    # Druckausgabe vorbereiten
        for i in data:
            datastr += "{:>02X}".format(i)
            cmd.append(i)

        l, c = self.__check_sum(cmd)      # Checksum und Länge bestimmen und anhängen
        cmd[2] = l                      # Länge im dritten Byte einfügen
        cmd.append(c)                   # checksumme anhängen

        if not self.ser == None:
            if self.readCTS and not self.ser.getCTS():
                time.sleep(0.01)
            self.ser.write(cmd)
            self.cnt_tx += 1

        # ggf. noch Ausgabe von Logdaten
        if (self.logToFile + self.logToTTY):
            t = "<- {} ".format(self.__format_time())
            for i in cmd:
                t+= "{:>02X} ".format(i)

            if self.logToFile and self.filehandle is not None:
                self.filehandle.write(t + "\n")

            if self.logToTTY:
                print(t)

        if setFrameId:
            self.X_jobs.addJob(self.frameId17, 0x97, dev, rxHandler)


# --- FRAME0x10 --- ZigBee Transmit Request ------------------------------------
    def sendDataFrame(self,
                dev,                         # erforderlich
                data,                        # erforderlich, ohne Daten sinnfrei
                rxHandler        = None,
                setFrameId       = True,     # optional, wenn <> none frameId = 0
                BroadcastRadius  = 0,        # optional, default (0 = maximum)
                Options          = 0         # optional, default 0
                ):
        # Rückgabe: keine
        if rxHandler is None:
            rxHandler = self.__doNix

        cmd = bytearray([0x7E, 0, 0, 0x10])   # Startdelimiter, Länge, Frame-Typ

        if setFrameId:
            self.frameId10 += 1
            if self.frameId10 > 255:
                self.frameId10 = 1
            cmd.append(self.frameId10)

        else:
            cmd.append(0)

        # das Ziel-Device kann über den Index oder den Namen angegeben werden
        if type(dev) is int:
            adr = self.X_adr.getStation(idx = dev)[1:3]
        # liefert auch den Index zurück
        elif type(dev) is str:
            tmp = self.X_adr.getStation(name = dev)
            dev = tmp[0]
            adr = tmp[1:3]
        else:
            print("Error type dev")

        # den Counter für gesendetete Frames erhöhen
        self.X_adr.incrTxCnt(dev)

        adr_64bit = ""
        for i in range(8):              # die 64-bit Destination Adresse
            adr_64bit += "{:>02X}".format(adr[0][i])
            cmd.append(adr[0][i])

        adr_16bit = ""
        for i in range(2):              # die 16-bit Destination Network Adresse
            adr_16bit += "{:>02X}".format(adr[1][i])
            cmd.append(adr[1][i])

        cmd.append(BroadcastRadius)     # Broadcast Radius
        cmd.append(Options)             # Options

        datastr = ""
        for i in data:      # die Parameter
            datastr += "{:>02X}".format(i)
            cmd.append(i)

        l, c = self.__check_sum(cmd)    # Checksum und Länge bestimmen und anhängen
        cmd[2] = l                      # Länge einfügen
        cmd.append(c)                   # checksumme anhängen

        if not self.ser == None:
            if self.readCTS and not self.ser.getCTS():
                time.sleep(0.01)
            self.ser.write(cmd)          # senden
            self.cnt_tx += 1

        # ggf. noch Logdaten ausgeben
        if self.logToFile or self.logToTTY:
            t = "<- {} ".format(self.__format_time())
            for i in cmd:
                t+= "{:>02X} ".format(i)

            if self.logToFile and self.filehandle is not None:
                self.filehandle.write(t + "\n")

            if self.logToTTY:
                print(t)

        if setFrameId:
            self.X_jobs.addJob(self.frameId10, 0x8B, dev, rxHandler)


    def __format_time(self):
        """
        Einheitliche Formatierung für die Sendezeit ggf. incl. Nachkommastellen
        """
        t = time.ctime().split()[3]
        if self.tm_dez:
            t += str(round(time.time() % 1, 1))[1:]
        return t


    def __check_sum(self, cmd):
        """
        liefert die Länge des Frame (incl. Prüfsumme) sowie
        die Prüfsumme für den Frame:
        Das erste Zeichen 0x7E und die beiden Längenbytes werden ausgelassen.
        Die Summe aller anderen Bytes (ab Byte 3) wird von 255 subtrahiert
        Die Summe ist vom Typ int, muss durch % 256 auf ein Byte reduziert
        werden.
        """
        sum = 0
        l = len(cmd)
        for i in range(3, l):
            sum += cmd[i]
        return ( l - 3, (255 - (sum % 256)))


    def __sendATDB(self):
        """ Sendet "verdeckt" (ohne Ausgabe im Log) das AT-Command "DB" um die
            Signalstärke des letzten empfangenen Frames auszulesen.
            Führt eine eigene Frame_Id mit, damit auch bei der Grundeinstellung
            Frame_ID == 0 das Modem abgefragt werden kann.
        """
        self.frameIdDB += 1
        if self.frameIdDB > 255:
            self.frameIdDB = 1

        cmd = [ 0x7E, 0, 4, 8, self.frameIdDB, ord("D"), ord("B"), 0]
        l, c = self.__check_sum(cmd)     # Checksum und Länge bestimmen

        cmd[7] = c
        # cmd[2] = l                    # ist immer '4'

        if not self.ser == None:
            self.ser.write(cmd)
            return (self.frameIdDB)

        else:
            return 0


    def __openLogFile(self, filename):
        """ Öffnet eine Logdatei mit dem Name filename
            oder legt einen Dateinamen selbst an,
            wenn der filename = ""         """
        if len(filename) == 0:
            # Dateinamen aus dem numerischen Zeitstring bilden und ".log" anhängen
            filename = str(int(time.time())) + ".log"

        print("open file: ", filename)
        file = open(filename, "w")
        file.write("open(): " +
                        time.strftime("%d.%m.%Y %H:%M:%S") + "\n")
        return file


    def closeLogFile(self):
        """ Schließt eine geöffnete Logdatei """
        if self.filehandle is not None:
            self.filehandle.write("close(): " +
                        time.strftime("%d.%m.%Y %H:%M:%S") + "\n")
            self.filehandle.close()
            self.filehandle = None
        self.logToFile = False
        self.X_adr.config(logToFile = false)


    def config(self, logToTTY = None, logToFile = None, readRSSI = None):
        """ Erlaubt die Veränderung von Einstellungen """
        if type(logToTTY) == bool:
            self.logToTTY = logToTTY
            self.X_adr.config(logToTTY = logToTTY)

        if type(logToFile) == bool:
            self.logToFile = logToFile
            self.X_adr.config(logToFile = logToFile)

        if type(readRSSI) == bool:
            self.readRSSI = readRSSI


    def cleanJobs(self, delay = 180):
        """ Löcht alle Einträge in der Jobliste, die älter als delay Sekunden
            sind. Die Zähler für fehlerhafte Übertragungen werden aktualisiert.
            Rückgabe : keine
        """
        x = self.X_jobs.cleanJobs(delay)
        if len(x):
            for idx in x:
                if idx >= 0:                    # -1 = local AT-Cmd oder Error
                    self.cnt_err += 1           # Fehler allgemein zählen
                    self.X_adr.incrErrCnt(idx)  # Fehler stationsabhängig eintragen

    def getJobsByStation(self, station):
        """ Gibt zurück, wieviele Rückmeldungen von der angegebenen Station
            erwartet werden -
            und vor wieviele Sekunden auf die älteste Anfrage abgechickt wurde.
            -> tuple: (Anzahl anstehender Frames, max Wartezeit)
        """
        return self.X_jobs.getJobsByStation(station)


    def getStation(self, idx = None, name = None, adr64 = None, adr16 = None):
        """ die Anfrage wird weitergereicht an X_adr() """
        return self.X_adr.getStation(idx, name, adr64, adr16)


    def readMyAdr(self):
        """ Sendet die localen At-Commands SL / SH,
            um zuerst die 64-Bit Seriennummer und danach
            den eigenen Namen zu ermitteln.
            Anschließend wird noch ein Remote At-Command an die eigene
            Station geschickt und dadurch die 16-Netzwerkadresse festgestellt.
            Wenn readRSSI = True, wird auch die Empfangsstärke abgeholt (-00db).
        """
        self.sendLocalAtCmd(atcmd = "SL", rxHandler = self.__readMyAdr)
        self.sendLocalAtCmd(atcmd = "SH", rxHandler = self.__readMyAdr)


    def __readMyAdr(self, station, frame):
        """ Liest mittels AT-Commands die eigene Seriennummer aus.
            Fügt die Antworten auf die lokalen AT-Commands SH / SL
            zu einer 64-Bit Adresse,
            prüft deren Existenz in der Stationsliste und
            gibt den Index der Station zurück,
            ermittelt über den Index den Namen der Station.
            Anschießend wird self.adr_self_idx in Xbee_adr()
            mit dem Index belegt, so dass die angeschlossene
            Station in der Stationsliste markiert werden kann. """

        #print("__readMyAdr():", frame)
        data = frame[-4:]           # die Daten stehen in den letzten 4 Bytes
        # die Adresse ist als Bytestring der Länge 8 vorbelegt.
        if frame[5:7] == bytes(("SH").encode('ascii')):
            self.adr64bit = data + self.adr64bit[4:]

        elif frame[5:7] == bytes(("SL").encode('ascii')):
            self.adr64bit = self.adr64bit[:4] + data

        tmp = self.X_adr.getStation(adr64 = self.adr64bit)
        idx = tmp[0]
        self.name = tmp[3]
        self.X_adr.setSelfIndex(idx)

        # Jetzt wird noch die 16-Bit Netzwerkadresse eingelesen
        # und dabei die Stationsliste aktualisiert
        # dazu wird ein beliebiges RemoteAtCmd an die eigende Adresse geschickt.
        # Im empfangenen Frame ist die 16-Bit Adresse enthalten.
        # Ausserdem wird die Empfangsstärke gemessen - in diesem Fall -00db
        if idx > -1:
            self.id = idx
            self.sendRemoteAtCmd(idx, "D0", rxHandler = self.__doGarNix)


# --- [ eof ] ------------------------------------------------------------------
