# -*- coding: UTF-8 -*-
#       10.11.2013
# --- [ xbee_adr.py ] ----------------------------------------------------------
#
# Verwaltet die Daten der Stationen (Adressen, Namen)
# Die einzelnen Stationen werden über ihren Index in der Stationsliste adressiert.
#
# Überprüft beim Empfang von Daten, ob die übertragene 16-Bit Adresse des Senders
# mit der gespeicherten Adresse in der Stationsliste übereinstimmt.
#
# Liest beim Programmstart die 64-/16-Bit Adressen und Stationsnamen aus der
# Datei "xbee.cfg" aus

class Xbee_adr(object):
    """ Legt ein Objekt zur Verwaltung der Stationen an.

        Liest die 64Bit Adressen und die Namen der Stationen aus einer
        Konfigurationsdatei aus.
        Die 16Bit Adressen werden zunächst mit FFFE initialisert und später nach dem
        ersten Kontakt mit der jeweiligen Station mit der vergebenen Adresse aktualisiert.
    """

    def __init__(self, configfile = "xbee.cfg", filehandle = None,
                                    logToFile = False, logToTTY = False):
        self.configfile = configfile
        # Hier werden die Default-Daten für Coordinator und Broadcast festgelegt.
        self.adr_64bit = [bytearray([0,0,0,0,0,0,0,0]), bytearray([0,0,0,0,0,0,0xFF,0xFF])]
        self.adr_16bit = [bytearray([0,0]), bytearray([0xFF, 0xFE])]
        self.adr_name  = ["COORDINATOR", "BROADCAST"]
        self.adr_rssi  = ["-", "-"]
        self.adr_rx_cnt  = [0,0]
        self.adr_tx_cnt  = [0,0]
        self.adr_err_cnt = [0,0]
        self.adr_self_idx = -1           # Index der eigenen Station in der List
        self.filehandle = filehandle
        self.logToFile  = logToFile
        self.logToTTY   = logToTTY

        # Die Daten der Stationen einlesen
        cnt = 0
        try:
            f = open(self.configfile)

            nw_adr = bytearray([0xFF, 0xFE])
            adr_64bit = list()
            adr_16bit = list()
            adr_names = list()

            zeile = f.readline()
            # in allen Zeilen mit Seriennummern werden 9 Kommata erwartet !
            while (zeile.count(',') == 9):
                zeile = zeile.split(',')
                l = bytearray()

                for i in range(0,8):
                    l.append(int(zeile[i],16))

                self.adr_64bit.append(bytes(l))
                self.adr_16bit.append(nw_adr)

                name = zeile[8]
                if len(name) == 0:
                    name = "#" + str(cnt + 1)
                self.adr_name.append(name)
                self.adr_rssi.append('?')
                self.adr_rx_cnt.append(0)
                self.adr_tx_cnt.append(0)
                self.adr_err_cnt.append(0)

                cnt += 1
                zeile = f.readline()

            f.close()

        except IOError:
            print("Datei {} nicht gefunden !". format(self.configfile))
            exit()

        print("importiert wurden {} Stationsadressen\n".format(cnt))


    def printStationList(self, station = -1):
        """ Stellt eine formatierte Liste der Stationen zusammen mit
            - 64-Bit-Adresse,
            - aktueller 16-Bit-Adresse,
            - StationsNamen
            - letztem verzeichneten RSSI-Wert
            - Anzahl der ausgeführten Übertragungen
            und druckt sie aus.
            Wird als Parameter der Index oder der Name einer Station übergeben,
            dann werden die Daten nur dieser Station als Tuple zurückgeliefert.
        """
        if station >= 0:
            # Liefert die Daten einer einzelnen Station als Tuple aus
            if type(station) is str:
                idx = -1
                for i in range(len(self.adr_64bit)):
                    if self.adr_name[i] == station:
                        idx = i
                        break
                if idx == -1:
                    s = "Unbekannte Station: {}".format(station)
                    print(s)
                    return

                station = idx

            s = (self.adr_64bit[station],
                 self.adr_16bit[station],
                 self.adr_name[station],
                 self.adr_rssi[station],
                 self.adr_rx_cnt[station],
                 self.adr_tx_cnt[station],
                 self.adr_err_cnt[station]
                )

            if self.filehandle is not None:
                print(s + "\n")
            if self.logToTTY:
                print(s)
            return s

        stationlist = list()
        stationlist.append("Idx 64-Bit-Adresse ------------ 16-Bit "\
                            "-- Name -------------- RSSI ------ RX ---- TX --- Err")
        for i in range(0, len(self.adr_64bit), 1):
            zeile = "{:<4}".format(i)

            for j in range(8):
                zeile += "{:>02X}".format(self.adr_64bit[i][j])
                zeile += " "

            if i == self.adr_self_idx:
                trennzeichen = "*** "
            else:
                trennzeichen = "--- "
            zeile += trennzeichen
            for j in range(2):
                zeile += "{:>02X} ".format(self.adr_16bit[i][j])
            zeile += trennzeichen
            zeile += "{:<20}".format(self.adr_name[i])

            zeile += "{:6}".format(self.adr_rssi[i])
            zeile += "{:8}".format(self.adr_rx_cnt[i])
            zeile += "{:8}".format(self.adr_tx_cnt[i])
            zeile += "{:8}".format(self.adr_err_cnt[i])
            stationlist.append(zeile)

        for z in stationlist:
            if self.filehandle is not None:
                print(z + "\n")
            if self.logToTTY:
                print(z)


    def checkAdr64(self, adr64, adr16):
        """ Prüft, ob die 64-Bit-Adresse in der StationsListe enthalten ist,
            wenn ja, dann wird geprüft, ob die 16-Bit-Netzwerkadresse
            mit der gespeicherten Adresse übereinstimmt,
            wenn nicht, dann wird die 16-Bit-Adresse aktualisiert.
            Wenn eine bestehende Verbindung unterbrochen wird, dann wird
            die 16-Bit-Adresse auf 'FFFE' geändert.

            Rückgabe:   ( 0, ) -> 16-Bit-Adresse unverändert
                        ( 1, "Hinweistext") -> Adresse geändert
                        (-1, "Hinweistext") -> 64-Bit-Adresse nicht vorhanden
        """

        result = -1

        # die Default-Adressen für Coordinator und Broadcast nicht prüfen
        # beginnen an Position 2 in der Liste
        for i in range(2, len(self.adr_64bit)):
            #print(self.adr_64bit[i] == adr64)
            # Die Seriennummer des Senders ist in der Liste eingetragen
            if self.adr_64bit[i] == adr64:

                # wenn die 16-Bit-Adresse des Senders nicht mit dem Listeneintrag
                # übereinstimmt, dann den Eintrag ändern
                #print(self.adr_16bit[i] != adr16)
                if self.adr_16bit[i] != adr16:
                    # die Adressen zur Dokumentation als String speichern
                    oldadr = ""
                    newadr = ""
                    adr_64bit_str = ""

                    for j in range(2):
                        oldadr += "{:>02X}".format(self.adr_16bit[i][j])
                        newadr += "{:>02X}".format(adr16[j])

                    # den neuen Wert in die Stationsliste eintrage
                    self.adr_16bit[i] = adr16

                    # die 64-Bit Adresse in einen String umwandeln
                    for j in range(8):
                        adr_64bit_str += "{:>02X}".format(adr64[j])

                    text = "   INFO : " + adr_64bit_str + " 16-Bit Adr changed: " \
                                     + oldadr + " -> " + newadr
                    result = 1
                    break

                else:
                    result = 0      # 0 = Adresse unverändert
                    break

        if result == -1:
            text = "   INFO : Unbekannte Adresse"

        if not result is 0:     # keine Veränderung
            if self.filehandle is not None:
                self.filehandle.write(text + "\n")
            if self.logToTTY:
                print(text)

        if i < len(self.adr_64bit):
            return i
        else:
            return -1
        #return (result, text)


    def getStation(self, idx = None, name = None, adr64 = None, adr16 = None):
        """ liefert die Daten einer Station zurück
            wobei nach Index, Name, 64Bit-Adresse oder 16Bit-Adresse gesucht
            werden kann.
            Rückgabe: tuple (idx, 64-Bit-Adresse, 16-Bit-Adresse, Name)
        """
        if type(idx) == int:
            if idx >= -1 and idx < len(self.adr_64bit):
                return (idx, self.adr_64bit[idx], self.adr_16bit[idx], self.adr_name[idx])

        elif type(name) == str:
            for i in range(len(self.adr_64bit)):
                if self.adr_name[i] == name:
                    return (i, self.adr_64bit[i], self.adr_16bit[i],name)

        elif type(adr64) == bytes or type(adr64) == bytearray:
            for i in range(len(self.adr_64bit)):
                if self.adr_64bit[i] == adr64:
                    return (i, self.adr_64bit[i], self.adr_16bit[i],self.adr_name[i])

        elif type(adr16) == bytes or type(adr16) == bytearray:
            for i in range(len(self.adr_64bit)):
                if self.adr_16bit[i] == adr16:
                    return (i, self.adr_64bit[i], self.adr_16bit[i],self.adr_name[i])

        return (-1, None, None , "?")


    def setSelfIndex(self, idx):
        """ Schreibt den Index der eigenen Station in Variable self._adr_self_idx
        """
        if idx > 1 and idx < len(self.adr_64bit):
            self.adr_self_idx = idx

    def incrRxCnt(self, idx):
        """ Incrementiert den Zähler für empfangene Sendungen """
        self.adr_rx_cnt[idx] += 1


    def incrTxCnt(self, idx):
        """ Incrementiert den Zähler für versandte Sendungen """
        self.adr_tx_cnt[idx] += 1


    def incrErrCnt(self, idx):
        """ Incrementiert den Zähler für aufgetretene Übertragungsfehler """
        self.adr_err_cnt[idx] += 1


    def setATDB(self, station, db):
        """ Schreibt den Wert der Signalstärke in die Stationsliste """
        self.adr_rssi[station] = "-{:>02}db ".format(db)


    def config(self, logToFile = None, logToTTY = None):
        if type(logToFile) is bool:
            self.logToFile = logToFile

        if type(logToTTY) is bool:
            self.logToTTY = logToTTY


if __name__ == "__main__":

    x = Xbee_adr(logToTTY = True)

    s64 = bytearray([0,0x13,0xA2,0,0x40,0x62,0x8D,0x12])
    s16 = bytearray([1,1])

    print("Suche nach Seriennummer: ", x.checkAdr64(s64, s16))

    s64 = bytearray([0,0x13,0xA2,0,0x40,0x62,0x8D,0x11])
    s16 = bytearray([1,2])

    print(x.getStation())
    print(x.getStation(idx = 2))
    print(x.getStation(name = "Station 1"))
    print(x.getStation(adr64 = s64))
    print(x.getStation(adr16 = s16))
##    x.getStationList()
##    x.getStationList(3)
##    x.getStationList("Station 2")
##    x.getStationList("Station 44")

    exit()

# --- [ eof ] ------------------------------------------------------------------
