# -*- coding: UTF-8 -*-
#       25.11.2013
# --- [ xbee_ctrl.py ] ---------------------------------------------------------
#
# Beim Senden der Frame 0x08, 0x10, 0x17 kann eine Funktion mitgegeben werden,
# die beim Entreffen der Rückmeldung ausgeführt werden soll.
# Wird keine Funktion angegeben, dann wird doNothing() eingesetzt´.
#
# Dieses Objekt führt eine Liste aller ausstehenden Antworten mit den
# - erwarteten FrameIds
# - dem Frametyp
# - der zurücksendenden Station
# - der auszuführenden Funktion
# - dem Zeitstempel beim Versenden.
#
# Beim Eingang von Rückmeldung auf die Frames 0x08, 0x10 und 0x17,
# - das sind die Frames 0x88, 0x8B, 0x97 -
# wird die Liste überprüft,
# ob dort ein Eintrag mit dem aktuellen Frametyp und der aktuellen FrameId vorliegt.
# Wenn ja, dann wird die zugeordnete Funktion ausgeführt und ihr die ID der
# sendenden Station und der empfangene Datensatz übergeben.
#
# Anhand der Liste kann kontrolliert werden, welche Antworten noch nicht
# eingegeangen sind.

import time

class Xbee_jobs(object):
    def __init__(self):
        self.jobs = []
        self.len_jobs = 0

    def addJob(self, frameId, frametyp, dev, function ):
        """ Fügt einen Job in die Liste ein
            Daten des zu empfangenden Frames:
            - FrameId,
            - FrameTyp,
            - auszuführende Funktion,
            - Zieldevice
            - Zeitpunkt des Versendens (um veraltete Eintrag auszusortieren)
        """
        self.jobs.append( (frameId, frametyp, function, dev, int(time.time())) )
        self.len_jobs = len(self.jobs)


    def checkJobs(self, frameId, frametyp):
        """ Prüft, ob für den eingegangenen Frame in der Jobliste
            ein Eintrag vorliegt.
            Liegen mehrere passende Einträge vor, dann werden alle gelöscht
            und -1 zurückgegeben.

            Übergabe: FrameId, FrameTyp

            Rückgabe:
            - Index in der Jobliste
            -1 wenn kein Eintrag vorhanden ist.
        """

        # nur ausführen, wenn tatsächlich Jobs eingetragen sind
        if self.len_jobs > 0:
            idx = -1
            cnt =  0
            for i in range(len(self.jobs)):
                # prüfen, ob FrameId etc. übereinstimmt
                if frameId == self.jobs[i][0]:
                    if frametyp == self.jobs[i][1]:
                        idx  = i
                        cnt += 1 # Counter um Mehrfacheinträge zu erkennen
        else:
            return -1

        # für den unwahrscheinlichen Fall, dass mehrere  Datensätze mit der
        # gleichen FrameId und Frametyp vorliegen, alle Datensätze löschen und
        # idx -1 zurückliefern
        if cnt > 1:
            idx = -1
            while cnt > 0:
                for i in range(len(self.jobs)):
                    if frameId == self.jobs[i][0]:
                        if frametyp == self.jobs[i][1]:
                            d = i
                            time.sleep(.5)
                            cnt -= 1
                            break
                # referenzierte Funktion ausführen mit den Parametern:
                # Station, (0xA5, FrameId, Frametyp)
                self.jobs[d][2]( self.jobs[d][3], (0xA5, self.jobs[d][0],
                                                              self.jobs[d][1]) )

                self.__writeErrLog("mehrfacher Eintrag", self.jobs[d])
                del self.jobs[d]

            time.sleep(.1)
        return idx


    def doJob(self, idx, frame):

        """ Es lag ein übereinstimmender Eintrag vor.
            Führt die hinterlegte Funktion aus und fügt die Id der
            sendenden Station sowie die Daten als Parameter ein.

            Rückgabe -> keine
        """
        # referenzierte Funktion ausführen, als Parameter die StationsId
        # aus der Jobliste und frame übergeben
        self.jobs[idx][2](self.jobs[idx][3], frame)
        del self.jobs[idx]
        self.len_jobs = len(self.jobs)


    def getJobsByStation(self, station):
        """ Durchsucht die Jobliste nach Sendungen, die an die genannte
            Station adressiert sind.
            Liefert die Anzahl und die maximale Wartezeit zurück.
            Je höher die Anzahl der ausstehenden Antworten und
            je größer die Wartezeit, umso wahrscheinlicher
            ist der Ausfall der Station.
        """
        t     = int(time.time())
        cnt   = 0
        max_t = 0

        for i in range(len(self.jobs)):
            if self.jobs[i][3] == station:
                cnt += 1
                w = t - self.jobs[i][4]
                if w > max_t:               # maximale Wartezeit
                    max_t = w

        return (cnt, max_t)


    def cleanJobs(self, delay = 180, enddevices = [] ):
        """ Löscht alle Jobs, die älter als delay Sekunden sind
            und deren Device-Nr nicht in die Liste der
            Enddevices eingetragen ist.

            Rückgabe:
            - Anzahl der gelöschten Einrträge,
            - Liste mit den zugehörigen Stationen
              wird benutzt, um die Fehlercounter der Stationen zu aktualisieren
        """
        #time.sleep(3)
        stations = []                  # für Fehlerliste die Stationen notieren
        cnt = 0
        t = int(time.time())
        while True:
            d = -1
            for i in range(len(self.jobs)):
                # wenn die Wartezeit überschritten ist
                if (t - self.jobs[i][4]) >= delay:
                    # und die Station nicht als Enddevice definiert ist
                    if not self.jobs[i][3] in enddevices:
                        d = i
                        break

            if d == -1:
                break

            else:
                cnt += 1

                self.__writeErrLog("cleanJobs()", self.jobs[d])
                # referenzierte Funktion ausführen mit den Parametern:
                # station, (0x5A, FrameId, Frametyp)
                self.jobs[d][2](self.jobs[d][3], (0x5A, self.jobs[d][0],
                                                            self.jobs[d][1]) )
                stations.append(self.jobs[i][3])   # Stations Idx
                del self.jobs[d]
        if cnt:
            self.__writeErrLog("cleanJobs()", "Anzahl gelöscht: {},\
                        Stationsliste: {}".format(cnt, stations))
        return stations


    def printJobs(self):
        """ Druckt zum Debuggen eine Liste aller Jobs """
        print("printJobs(): Liste enthält {} Einträge".format(len(self.jobs)))
        for i in self.jobs:
            print("Job: FrameId {}, Frametyp {}, Dev {}, Function {}".format(
                                         i[0], "{:02X}".format(i[1]), i[3], i[2] ))


    def __writeErrLog(self, msg, data):
        """ Schreibt eine Fehlermeldung in das File error.log """
        f = open("error.log", "a")
        out = time.strftime("%d.%m.%Y %H:%M:%M - ")
        out += "{}: {}\n".format(msg, data)
        f.write(out)
        f.close()

        print(out)


if __name__ == "__main__":

    # Zum Testen muss ein Delay in cleanJobs() auskommentiert werden !!!!!!
    import xbee_adr

    def printjobs( station, frame ):
        print("printjob() mit Parameter {} {} aufgerufen".format(station, frame))

    xadr = xbee_adr.Xbee_adr()
    xjob = Xbee_jobs()

    xjob.addJob(frameId = 4, frametyp = 0x88, dev = 1, function = printjobs )
    xjob.addJob(frameId = 4, frametyp = 0x88, dev = 2, function = printjobs )
    xjob.addJob(frameId = 2, frametyp = 0x97, dev = 3, function = printjobs )
    xjob.addJob(frameId = 1, frametyp = 0x8B, dev = 0, function = printjobs )

    r = xjob.checkJobs(frameId = 4, frametyp = 0x88)
    print(r)
    if r:
        xjob.doJob(r, (1,2,3,4))
    xjob.cleanJobs(3, enddevices = [])
    #xjob.cleanJobs(3, enddevices = (1,2))
    exit()
