# -*- coding: UTF-8 -*-
#       24.11.2013
# --- [ xbee_rx.py ] -----------------------------------------------------------
#
# Ist für das Empfangen von Frames via Seriellen Schnittstelle zuständig
# Prüft die Checksumme und liefert zurück
# - den Frame ohne Header und Prüfsumme
# - die Zeit des Empfangs

import serial
import threading
import time

class Xbee_rx(threading.Thread):
    def __init__(self,
                ser,
                callback = None
                ):
        threading.Thread.__init__(self)
        self.ser = ser
        self.cb = callback
        self.threadend = False
        self.framecnt = 0
        self.show_frame = 0x200         # ein Flag welches angibt, ob automatisch die
                                        # Signalstärke ausgelesen wird

    def exit(self):
        """ Beendet den Thread (mit kurzer Verzögerung durch die sleep's) """
        self.threadend = True


    def run(self):
        """
        Wenn mindestens 3 Byte an Daten anstehen, dann werden diese eingelesen
        Es sollte sich um die ersten drei Bytes eines Frames halten:
        Daher mit 0x7E und 0x00 beginnen, nachfolgend das Byte mit der Länges des Datensatzes.
        Das muss gelesen werden, um die Anzahl der nachfolgenden Bytes des Frames
        zu kennen.
        (Laut Datenblatt legen Byte 1 und Byte 2 die Länge fest.
        Allerdings wird ein Frame nie die Länge von 255 Bytes erreichen,
        Byte 1 wird also immer 0 sein.)
        Danach wird auf den Eingang des Restes des Datensatzes gewartet.
        Ist er komplett, dann wird er eingelesen und die Prüfsumme getestet.
        Ist diese korrekt, dann kann der Datensatz per Callback übergeben werden

        Rückgabe:
        - Datenframe als Bytestring ohne Prüfsumme,
        - Zeitpunkt es Empfangs als String.
        """
        self.ser.flush()                        # alle anstehenden Daten löschen

        while self.threadend is False:
            i = self.ser.inWaiting()

            if i is 0:
                time.sleep(0.01)                # 10ms warten

            else:
                # wird nur von readRSSI genutzt um zu erkennen, ob neue Frames
                # eingegangen sind
                self.framecnt = (self.framecnt + 1) & 0xFF

                # die Systemzeit mit 1 Nachkommastelle festhalten
                timestring = (time.ctime()[11:19]+ "." +
                                                str(round(time.time(),1))[-1:])
                while i < 3:
                     time.sleep(0.01)                # 10ms warten
                     i = self.ser.inWaiting()

                frm_head = self.ser.read(3)  # die ersten 3 Bytes einlesen

                # Wenn die ersten beiden Bytes nicht mit dem erwarteten Beginn eines Frames übereinstimmmen.
                # dann wurde kein Framestart erkannt, wir sind mitten in einem Frame gelandet.
                # Warten bis zum Ende der Übertragung und alle Daten in der Zwischenzeit verwerfen.

                if ( not (frm_head[0] == 0x7E and frm_head[1] == 0)):
                    while True:
                        self.ser.flush()
                        time.sleep(.005)    # 5 Millisekunden, sollte für mindestens 1 Zeichen ausreichen
                        if self.ser.inWaiting() == 0: # wenn kein Byte eingegangen,
                            break           # dann kann die Schleife verlassen werden

                # der verbleibende Rest des Frames mit den Daten einlesen
                # die Anzahl steht in (Byte 1 * 256) + Byte 2
                # wobei Byte 1 immer '0' sein wird, also nicht ausgewertet wird ...
                data_len = frm_head[2] + 1

                # Warten bis das Frame vollständig im Empfangspuffer vorliegt
                while (self.ser.inWaiting() < data_len):
                    time.sleep(.005)

                frm_data = self.ser.read(data_len)

                # Jetzt ist das komplette Frame eingelesen, nun die Prüfsummme testen
                # Die Summe des Datenbereiches bis incl. der Prüfsumme muss 255 ergeben
                chk = 0
                for i in frm_data:
                    chk += i

                if ((chk % 256) != 255):    # dann stimmt die Checksumme nicht !

                    #raise ValueError("Frame Checksum Error")

                    # Ein Fehler in der Checksum ist aufgetreten
                    # die fehlerhaften Daten ausgeben
                    print("Frame Checksum Error : ", end = "")

                    hex_data=""
                    for i in range(len(frm_head)):
                        hex_data += "{:>02X}".format(frm_head[i])
                    print(hex_data, end = " ")

                    hex_data=""
                    for i in range(len(frm_data) - 1):
                        hex_data += "{:>02X}".format(frm_data[i])
                    print(hex_data)

                    # falls noch Daten empfangen werden, alles verwerfen
                    # bis eine Sendepause von 10ms erkannt wird
                    while True:
                        self.ser.read(self.ser.inWaiting())
                        time.sleep(0.01)
                        if self.ser.inWaiting() == 0:
                            break
                    continue

                # Die Checksumme war korrekt ----------------------------------

                # liefert die Daten als Bytearray ohne die Prüfsumme
                # und die Zeit formatiert als String zurück

                self.cb( frm_head + frm_data[:-1], timestring )
                continue

        print("thread closed")
        exit()

# --- [ eof ] ------------------------------------------------------------------
